'Modified version TO test OUT timing the parts more accurately
'changed to match the one oposted by Hutch in the masm forum so I can compare times

'this line only needed to enablethe accurate timing:
$INCLUDE "win32api.inc"



'http://www.powerbasic.com/support/forums/Forum12/HTML/001601.html
'P.Dixon , Sunderland, England
'mcm@pdixon.com
'compiled with PB/CC1.00
'This code is placed in the public domain by the author 26 March 2005
'
'I reckon this meets the specs except that I used PBCC1.0 (not 2.0+)
'
'Code overview
'The file is opened and read into a string in one go.
'It's then scanned from start to finish, one byte at a time, and all required processing is
'done on that byte.
'No byte is read twice so the whole file is scanned only once, there is never a need to reread
'a byte within the file.
'A straightforward algorithm is used to parse the data.
'A byte is either in the record header (1st 7 bytes of each record) or it isn't.
'If it's in the header, there are 4 possibilities:
'   1) its a new line (a blank line has no 7 byte header) so skip it and start on next record
'   2) The header starts with 4 spaces. This is quick to check and covers over half of the lines
'      and saves having to check for "book" which is slower. Treat this as in case 4
'   3) The header contains the word "book". Skip the remainder of this record
'   4) The header contains other stuff. Ignore the header and start processing the rest of the
'      record.
'
'If a byte is not in the header then we need to count/process it.
'First, a 256 byte look up table is used to do most of the work.
'This one look-up converts upper case to lower case,
'   converts all numbers/punctuation marks to spaces,
'   ignores the "'" which can be simply ignored to fit the rules,
'   and identifies end of record characters.
'
'   At this point we can conveniently count character substitutions by comparing the original
'   character with the looked-up character and, if different, incrementing a counter array
'   indexed by the original (replaced) character.
'
'
'Once converted, each character is parsed into words according to very simple rules.
'We are either within a word or between words.
'   If we are not yet in a word (no alphabetic character has been found) then:
'       If this is a valid character, it's the start of a new word.
'       If this is not a valid character then we're still between words so skip it.
'
'   If we are within a word then:
'       If this is a valid character it's the next letter of the word, append it.
'       If this is not a valid character then we've found the end of this new word.
'
'
'When a new word is found, we count it and then need to compare it with all previously found
'words to see if it's unique. The obvious way is to scan the array of already found words from
'start to finish, or until the word is found, but this is very inefficient as the number of
'unique words found becomes large.
'A better way is to use a sorted tree structure.
'The new word is compared with the word at root of the tree.
'   If it's empty, we've found a new word. Insert the new word in the tree and count it.
'
'   If the new word is equal to the one in the tree, we've found a duplicate word,
'   we can count it for extra points or just ignore it to fit the rules.
'
'   If the new word is > or < the one in the tree then we follow a >link or <link to the next
'   item in the tree.
'
'   We then repeat the process on the next tree location until we find a match or find that
'   it's a new word.
'
'This method hugely reduces the number of string comparisons needed but it can be improved further.
'
'A neat trick can be used here to help. If the first word happened to be "A" then the root of
'the tree would be "a" and all subsequent compares with the root (that's every time we get a
'new word) will be > the root. This is a waste of a compare for every new word. Instead, we
'can seed the tree with a middle-ranking-root which will give us a 50-50 split on the first
'compare. This increases the search efficiency and we just ignore the seed at the end.
'
'Since we're just counting unique numbers we don't need to search all items in the tree.
'Instead, we could create a tree with 256 roots rather than just 1. Which root we start from
'will be determined by the first letter in the new word.
'This way we have 26 separate trees (only 26 characters are valid so the remainder of the 256
'roots are never used) and we reduce the depth of the tree search by about 4 levels, halving
'the number of compares needed.
'This is a very efficient trick to save time at virtually no cost.
'Only 256*8 bytes are "wasted" in implementing it.
'
'We can take this approach 1 step further.
'If we use the first 2 letters of a new word then we can get 65536 roots (of which only about
'700 are used as there are limited numbers of 2 letter combinations in the entire 65536 range).
'This is feasible as, although it "wastes" a lot more memory (about 0.5MB) this is trivial on
'any modern computer. It reduces the number of comparisons by another 40% compared to 256 roots.
'
'String manipulations appear to be quite slow so pointers into blocks of data are used throughout
'for speed.
'When the file is fully scanned the tree of words is extracted into a standard string array
'where sorting or searching can be easily done if required.
'
'The total number of words (excluding headers and "book titles") is counted after each new word
'is found.
'The total number of unique words is counted when a new word finds a place in the tree.
'The number of character substitutions is counted immediately after the substitution is made.
'These make up the required output of the program.
'Further processing could be done on the data for extra points.
'We have arrays containing all unique words and how many times they occurred.

%maxwordlength=256      'maximum allowable characters in a single word
%maxworkspace=1000000   'workspace to reserve for words found,
                        '= (65536*8+enough to store all words + 7 bytes/word)..at least
                        'the test file uses 685,000 =  161,000 for words + 524,000 fixed.
%maxuniquewords=30000   'maximum number of unique words to allow. Test file has 12,685.


FUNCTION PBMAIN()
QueryPerformanceCounter (tc0&&)

'get the file name and other parameters
s$=TRIM$(COMMAND$)

'if none given on the command line then ask for them
IF s$="" THEN LINE INPUT "File path/name?",s$

'why not time it.
StartTime!=TIMER

'Does the user want the HELP screen?
IF INSTR(s$,"/?") THEN
   STDOUT "Perform file processing of test file according to:"
   STDOUT "http://www.powerbasic.com/support/forums/Forum12/HTML/001733.html"
   STDOUT "Author: P.Dixon"
   STDOUT
   STDOUT "Usage:
   STDOUT "CONTEST [[drive:][path]filename] [/?] [/n] [/1] [/a]"
   STDOUT
   STDOUT "  /?       Print this help file."
   STDOUT "  /n       Don't wait for key press at end."
   STDOUT "  /a       Sort word list alphabetically and append to output."
   STDOUT "  /1       Sort word list by number of occurrences and append to output."

   STDOUT "[[drive:][path]filename]"
   STDOUT "           Specifies the file to process."
   STDOUT "           If no file is given, the user will be prompted for it."
   GOTO xit
END IF

'check if the user wants to supress the "press a key to continue" prompt.
'this may be useful when the output is directed to a file or the program is shelled by another
IF INSTR(s$,"/n") OR INSTR(s$,"/N") THEN
    s$=REMOVE$(s$,"/n")
    s$=REMOVE$(s$,"/N")
    DontWait&=1
END IF

'Check if the user wants the full alphabetically sorted word list
IF INSTR(s$,"/a") OR INSTR(s$,"/A") THEN
    s$=REMOVE$(s$,"/a")
    s$=REMOVE$(s$,"/A")
    SortAlpha&=1
END IF

'Check if the user wants the full word list sorted on number of occurrences
IF INSTR(s$,"/1") THEN
    s$=REMOVE$(s$,"/1")
    SortNum&=1
END IF

'what's left should be the file to process
fn$=TRIM$(s$)
'but better remove quotes if they're there
fn$=TRIM$(fn$, CHR$(34))


'setup stuff
REGISTER atbp AS LONG, idx AS LONG
DIM wrd AS ASCIIZ*%maxwordlength  'the maximum length word allowed
DIM wrdp AS BYTE PTR, wrdcnt AS LONG,wrdptr AS LONG PTR
DIM lt AS STRING,ltp AS BYTE PTR
DIM b AS STRING , bp AS BYTE PTR
DIM NotInAWord AS LONG
DIM  hwm AS LONG, c AS LONG,header AS LONG, temp AS LONG
DIM GreaterThan AS LONG,LessThan AS LONG
DIM mp1 AS LONG PTR, mp2 AS LONG PTR, mp3 AS LONG PTR, wsp AS LONG PTR ,miscp AS LONG PTR
DIM iptr AS WORD PTR

'estimate size of arrays needed, for now just fix it at 30000 unique words max
DIM mx AS LONG, gotit AS LONG
mx=%maxuniquewords+65536:hwm=0  'enough for 65536 seeds and the max unique words
DIM words(%maxuniquewords) AS STRING, count(mx) AS LONG, upp(mx) AS LONG, dn(mx) AS LONG
DIM ws(%maxworkspace) AS BYTE  'my own workspace, string arrays are too slow
DIM WordAddr(mx) AS LONG
wsp=VARPTR(ws(0))
Maxwsp&=wsp+%maxworkspace

'point to the place where I'll assemble the individual words
wrdptr=VARPTR(wrd)
'rplace will count the replaced characters
DIM rplace(256) AS LONG


'assemble the 256 byte lookup table that'll do a lot of the work on individual characters
'if char=0 then convert to a space
'if char=1, then delete it
'else char= the required replacement for the original
lt=STRING$(39,0)+CHR$(0)+STRING$(8,0)+"0123456789"+STRING$(7,0)+"ABCDEFGHIJKLMNOPQRSTUVWXYZ"+STRING$(6,0)+_
   "abcdefghijklmnopqrstuvwxyz"+STRING$(133,0)
ltp=STRPTR(lt)


'use STDOUT throughout so output can be redirected to a file if needed.
STDOUT "Processing "+fn$
'get the file
'first, see if it exists..
ERRCLEAR
FileHandle&=FREEFILE
OPEN fn$ FOR INPUT AS fileHandle&
IF ERR THEN STDOUT "No such file.":GOTO errxit
CLOSE fileHandle&

QueryPerformanceCounter (tc1&&)


'it exists, now see if we can open it
ERRCLEAR
FileHandle&=FREEFILE
OPEN fn$ FOR BINARY AS fileHandle&
IF ERR THEN STDOUT "Can't open file.":GOTO errxit
chars&=LOF(FileHandle&)
GET$ FileHandle&,chars&,b
CLOSE fileHandle&

QueryPerformanceCounter (tc2&&)

'pointer to the character I'm processing
bp=STRPTR(b)
EndOfString&=bp+chars&
wrdp=wrdptr         'I need both byte and LONG pointers to the same address for speed later
hwmlimit&=65536     'point at first real word, above the 65536 seeds

'try seeding 65536 locations of the tree as described earlier
FOR r&=0 TO hwmlimit& -1
    count(r&)=1:WordAddr(r&)=wsp
    @wsp =&h676700ff 'this causes a roughly 50:50 split on the first compare
    @wsp[1]=0        'end of word marker
    wsp=wsp+8
NEXT

'hwm is the High Water Mark. It's the highest address used so far in my workspace indexes
'set it to the next free space after the seeding
hwm=hwmlimit&

Header=1:NotInAWord=1  'I'm In the header at the start and Not in a word.

'main loop start
LP:
atbp=@bp
    c=@ltp[atbp]

    INCR bp  'point to next character

    IF NotInAWord THEN
        IF c<2 THEN
            'c<2 means character is to be ignored...
            IF atbp=10 THEN header=1   'unless it's a LF then we are entering the next header

        ELSE
            'it's the 1st char of a new word
            @wrdp=c         'write first character to word
            wrdcnt=1        'we have 1 character in the word
            NotInAWord=0    'we are now in a word
        END IF

    ELSE  'NotInAWord
        'we must be in a word here
        IF c=0 THEN
            'end of a word, set end of string mark
            miscp = wrdp+wrdcnt
            @miscp=0      'write 8 zero bytes to guarantee a 32 bit dword=0 at the end
            @miscp[1]=0

            'PUT the new word into the arrays
            gotit=0       'this is the flag to show we found the location for the word in the tree
            iptr=wrdp     'we need to convert a byte pointer to an integer pointer to do the 65536 seeding
            idx=@iptr     'then the index to start searching at is the required root from the 65536 trees

            DO
                'compare the 2 words. Do quick check on first 4 characters. If not equal, we found it
                miscp=WordAddr(idx)
                temp=@miscp - @wrdptr

                IF temp>0 THEN
                    GOTO Greater
                ELSEIF temp<0 THEN
                    GOTO Lesser
                ELSE
                    'first 4 letters match so we need to scan the rest of the letters in the words as well
                    cnt&=0
                    DO
                        INCR cnt&
                        temp=@miscp[cnt&] - @wrdptr[cnt&]

                        IF temp>0 THEN
                            GOTO Greater
                        ELSEIF temp<0 THEN
                            GOTO Lesser
                        END IF

                    LOOP UNTIL  @wrdptr[cnt&] =0   'we end when we hit an end of word marker (=0)

                END IF
                'if we arrive here, they must be equal
equaler:
                gotit=1            'we found the word's tree location
                INCR count(idx)    'keep count of how many of this word we've found
                GOTO GotitLoop2    'and exit

greater:
                'the word is before the current idx, follow the down link
                IF dn(idx)=0 THEN
                    'not yet defined, so create it
                        dn(idx)=hwm         'hwm is the next free space in the storage area index
                        WordAddr(hwm)=wsp   'wsp is the pointer to the next free space in the workspace
                        miscp=wrdp
                        WHILE @miscp        'copy the new word to the workspace/tree
                            @wsp=@miscp
                            INCR wsp
                            INCR miscp
                        WEND

                        @wsp=0              'set an "end of word" marker (0 makes it an asciiz, useful later)
                        INCR wsp            'point to next free space in workspace

                        count(hwm)=1        'it's a new word, set count to 1
                        INCR hwm            'point to next free index location
                        gotit = 1           'flag that we're finished with this word

                        'to check for IDX or Workspace running out, check hwm and wsp here
                        'possibly, error out or redim preserve the relevant arrays to make them bigger
                        'making them too big to start with slows things down
                        IF hwm > 95536 OR wsp > Maxwsp& THEN STDOUT "Out of workspace":GOTO errxit

                ELSE
                    'follow it
                    idx=dn(idx)         'not found yet, follow the link DOWN to the next tree branch
                END IF
                GOTO GotitLoop         'iterate loop
lesser:
                'the word is after the current idx, follow the up link
                IF upp(idx)=0 THEN
                    'not yet defined, so create it
                    upp(idx)=hwm        'hwm is the next free space in the storage area index
                    WordAddr(hwm)=wsp   'wsp is the pointer to the next free space in the workspace
                    miscp=wrdp
                    WHILE @miscp        'copy the new word to the workspace/tree
                        @wsp=@miscp
                        INCR wsp
                        INCR miscp
                    WEND
                    @wsp=0              'set "end of word" marker (0 makes it an asciiz, useful later)
                    INCR wsp            'point to next free space in workspace

                    count(hwm)=1        'it's a new word, set count to 1
                    INCR hwm            'point to next free index location
                    gotit = 1           'flag that we're finished with this word

                    'to check for IDX or Workspace running out, check hwm and wsp here
                    'possibly, error out or redim preserve the relevant arrays to make them bigger
                    'making them too big to start with slows things down
                    IF hwm > 95536 OR wsp > Maxwsp& THEN STDOUT "Out of workspace":GOTO errxit

                ELSE
                    'follow it
                    idx=upp(idx)        'not found yet, follow the link UP to the next tree branch
                END IF
GotitLoop:
            LOOP UNTIL gotit            'end of the loop trying to insert a word in the tree
GotitLoop2:

            'Now switch back to not in a word
            NotInAWord=1
            INCR noofwords&             'count the number of words found

        ELSEIF c=1  THEN
            'ignore this character
            GOTO skip
        ELSE
            'add next character to new word
            @wrdp[wrdcnt]=c
            INCR wrdcnt        'and incr count of characters in the word

            ' to check for maximum wordlength being exceeded, trap it here (wrdcnt>256)
            IF wrdcnt >= %maxwordlength THEN STDOUT "Word to long":GOTO errxit

        END IF

    END IF   'NotInAWord


skip:

'compiler won't do fast compare of a pointer so assign the pointer value, then compare
tx&=bp
IF EndOfString& >= tx& THEN GOTO lp
' finished the main loop! all the work is done.

QueryPerformanceCounter (tc3&&)

DECR hwm        'point to last USED index

'got to time how long I took
EndTime!=TIMER
TotTime&=1000*(EndTime!-StartTime!)
IF TotTime& < 0 THEN TotTime&=TotTime&+86400000   'correct for midnight roll over


'now copy results to a string array, data is stored as asciiz strings anyway
'hwmlimit is the cut off for the seeds, we need to exclude those values
DIM azp AS ASCIIZ PTR
FOR r& = hwmlimit& TO hwm
    azp=WordAddr(r&)
    words(r&-hwmlimit&)=@azp
    count(r&-hwmlimit&)=count(r&)
NEXT

top&=hwm-hwmlimit&
REDIM PRESERVE words(0 TO top&),count(0 TO top&)

'print the required data
STDOUT
STDOUT "Total number of countable words="+STR$(noofwords&)
STDOUT "Unique countable words="+STR$(top&+1)

STDOUT
STDOUT "Done in "+STR$(TotTime&/1000)+"sec."
STDOUT

'check if the optional word lists are needed and sort into the required order
IF SortNum& THEN
   ARRAY SORT count() ,TAGARRAY words()
ELSEIF SortAlpha& THEN
   ARRAY SORT words() ,TAGARRAY count()
END IF

QueryPerformanceCounter (tc4&&)

QueryPerformanceFrequency (frq&&)

STDOUT "Input file="+STR$(INT((tc2&&-tc1&&)/frq&&*1000))+"ms"
STDOUT "Scan and list unique words="+STR$(INT((tc3&&-tc2&&)/frq&&*1000))+"ms"
STDOUT "sort words="+STR$(INT((tc4&&-tc3&&)/frq&&*1000))+"ms"
STDOUT "Total time="+STR$(INT((tc4&&-tc0&&)/frq&&*1000))+"ms"
STDOUT


'print the optional wordlists
IF SortNum& OR SortAlpha& THEN
   STDOUT "  Quantity Word"
FOR r&=0 TO top&
    STDOUT RIGHT$("          "+STR$(count&(r&)),10)+" "+words(r&)
    NEXT
END IF


errxit:
'wait for a key press unless specifically asked not to
IF DontWait&=0 THEN
   STDOUT "press a key to continue.";
   WAITKEY$
END IF

xit:


END FUNCTION
